# Created by Matyáš Pokorný on 2018-01-26.

module KXI
	module Application
		# Allows simple unstructured (time-independent) logging
		class Logger
			# Gets the message event
			# @return [KXI::Application::Event] Message event
			def on_message
				@event
			end
			
			# Attaches action to message event
			# @param min [KXI::Application::Logger::Message::Severity] Minimal severity of message
			# @yield [msg] Event action
			# @yieldparam msg [KXI::Application::Logger::Message] Logger message
			# @return [Number] Id of event hook
			def attach(min = Message::Severity::COMMAND, &block)
				return @event.add { |s, msg| block.call(msg) if msg.severity.level >= min.level }
			end
			
			# Instantiates the {KXI::Application::Logger} class
			def initialize
				@event = KXI::Application::Event.new
				@start = DateTime.now
			end
			
			# Logs message with the {KXI::Application::Logger::Message::Severity::COMMAND} severity
			# @param cmd [String] Executed command
			def command(cmd)
				log(Message::Severity::COMMAND, cmd)
			end
			
			# Logs message with the {KXI::Application::Logger::Message::Severity::TRACE} severity
			# @param msg [String] Text of message
			def trace(msg)
				log(Message::Severity::TRACE, msg)
			end
			
			# Logs message with the {KXI::Application::Logger::Message::Severity::DEBUG} severity
			# @param msg [String] Text of message
			def debug(msg)
				log(Message::Severity::DEBUG, msg)
			end
			
			# Logs message with the {KXI::Application::Logger::Message::Severity::INFO} severity
			# @param msg [String] Text of message
			def info(msg)
				log(Message::Severity::INFO, msg)
			end
			
			# Logs message with the {KXI::Application::Logger::Message::Severity::WARNING} severity
			# @param msg [String] Text of message
			def warning(msg)
				log(Message::Severity::WARNING, msg)
			end
			
			# Logs message with the {KXI::Application::Logger::Message::Severity::ERROR} severity
			# @param msg [String] Text of message
			def error(msg)
				log(Message::Severity::ERROR, msg)
			end
			
			# Logs message with the {KXI::Application::Logger::Message::Severity::FATAL} severity
			# @param msg [String] Text of message
			def fatal(msg)
				log(Message::Severity::FATAL, msg)
			end
			
			def log(severity, msg)
				now = DateTime.now
				@event.fire(self, Message.new(severity, now - @start, now, msg))
			end
			
			private :log
			
			# Represents a log message
			class Message
				# Represents severity of message
				class Severity
					# Gets the name of severity
					# @return [String] Name of severity
					def name
						@name
					end
					
					# Gets the level of severity
					# @return [Number] Level of severity
					def level
						@lvl
					end
					
					# Instantiates the [KXI::Application::Logger::Message::Severity] class
					# @param nm [String] Name of severity
					# @param lvl [Number] Level of severity
					def initialize(nm, lvl)
						@name = nm
						@lvl  = lvl
					end
					
					
					# Severity of messages that log execution of external commands
					COMMAND = Severity.new('command', 0)
					# Severity of messages that allow tracing of execution paths
					TRACE = Severity.new('trace', 1)
					# Severity of messages used for debugging
					DEBUG = Severity.new('debug', 2)
					# Severity of messages that user should know about
					INFO = Severity.new('info', 3)
					# Severity of messages that notify about something that might lead in future to problems
					WARNING = Severity.new('warning', 4)
					# Severity of messages that notify about problems that application can recover from
					ERROR = Severity.new('error', 5)
					# Severity of messages that notify about problems that application cannot recover from
					FATAL = Severity.new('fatal', 6)
				end
				
				# Gets the severity of message
				# @return [KXI::Application::Logger::Message:Severity] Severity of message
				def severity
					@sev
				end
				
				# Gets time at which the message was created
				# @return [DateTime] Time of creation
				def absolute_time
					@ttime
				end
				
				# Gets time, relative to creation of logger, at which the message was created
				# @return [DateTime] Relative time of creation
				def relative_time
					@rtime
				end
				
				# Gets the text message
				# @return [String] Message
				def message
					@msg
				end
				
				# Instantiates the {KXI::Application::Logger::Message} class
				# @param severity [KXI::Application::Logger::Message::Severity] Severity of message
				# @param rtime [DateTime] Relative time of message
				# @param ttime [DateTime] Absolute time of message
				# @param message [String] Text of message
				def initialize(severity, rtime, ttime, message)
					@sev   = severity
					@rtime = rtime
					@ttime = ttime
					@msg   = message
				end
			end
		end
	end
end