# Created by Matyáš Pokorný on 2017-12-28.

module KXI
	module Collections
		# Makes array enumerable
		class ArrayCollection < KXI::Collections::Enumerable
			# Instantiates the {KXI::Collections::ArrayCollection} class
			# @param array [Array] Array for enumeration
			def initialize(array = [])
				super()
				@arr = array
			end
			
			# Creates a new {KXI::Collections::Enumerator} bound to this instance
			# @return [KXI::Collections::Enumerator] Enumerator bound to this instance
			def create_enumerator
				ArrayEnumerator.new(@arr)
			end
			
			# Obtains value in array at specific index
			# @param index [Number] Index in array to obtain
			# @return [Object] Value in array at specified index
			# @raise [KXI::Exceptions::OutOfRangeException] Raised when given index is out of range of array
			def [](index)
				raise(KXI::Exceptions::OutOfRangeException.new(index)) if @arr.length == 0
				raise(KXI::Exceptions::OutOfRangeException.new(index, 0, @arr.length - 1)) if index < 0 or index >= @arr.length
				lock do
					return @arr[index]
				end
			end
			
			# Sets value of array at specific index
			# @param index [Number] Index in array to set
			# @param value [Object] Value to set
			# @return [Object] Set value
			# @raise [KXI::Exceptions::OutOfRangeException] Raised when given index is out of range of array
			# @raise [KXI::Exceptions::CollectionException] Raised when collection cannot be locked for writing
			def []=(index, value)
				raise(KXI::Exceptions::OutOfRangeException.new(index)) if @arr.length == 0
				raise(KXI::Exceptions::OutOfRangeException.new(index, 0, @arr.length - 1)) if index < 0 or index >= @arr.length
				lock(true) do
					@arr[index] = value
				end
				return value
			end
			
			# Adds value into array
			# @param value [Object] Value to set
			# @return [Number] Assigned index
			# @raise [KXI::Exceptions::CollectionException] Raised when collection cannot be locked for writing
			def add(value)
				idx = @arr.length
				lock(true) do
					@arr.push(value)
				end
				return idx
			end
			
			# Removes item at specific index
			# @param index [Number] Index to remove
			# @return [Object] Removed item
			# @raise [KXI::Exceptions::OutOfRangeException] Raised when given index is out of range of array
			# @raise [KXI::Exceptions::CollectionException] Raised when collection cannot be locked for writing
			def remove_at(index)
				raise(KXI::Exceptions::OutOfRangeException.new(index)) if @arr.length == 0
				raise(KXI::Exceptions::OutOfRangeException.new(index, 0, @arr.length - 1)) if index < 0 or index >= @arr.length
				ret = nil
				lock(true) do
					ret = @arr.delete_at(index)
				end
				return ret
			end
			
			# Enumerates array
			class ArrayEnumerator < KXI::Collections::Enumerator
				# Instantiates the {KXI::Collections::ArrayCollection::ArrayEnumerator} class
				# @param array [Array] Array for enumeration
				def initialize(array)
					@arr     = array
					@current = 0
				end
				
				# Selects first item in collection
				# @return [Bool] True if collection contains elements; otherwise false
				def rewind
					@current = 0
					return @arr.length > 0
				end
				
				# Advances enumerator to next item
				# @return [Bool] True if item is available; false otherwise
				def next
					@current = @current + 1
					return @current < @arr.length
				end
				
				# Returns current item
				# @return [Object] Current item
				def current
					@arr[@current]
				end
			end
			
			private :create_enumerator
		end
	end
end