# Created by Matyáš Pokorný on 2018-04-14.

module KXI
	module Math
		# Represents a vector
		class Vector
			# Returns the dimension of vector
			# @return [Integer] Dimension of vector
			def dimension
				@dim
			end
			
			# Instantiates the {KXI::Math::Vector} class
			# @param [Integer] dim Dimension of vector
			# @yield [d] Generator function
			# @yieldparam Dimension to generate
			# @yieldreturn [Numeric] Result of generator
			# @raise [KXI::Exceptions::InvalidTypeException] When generator function returns value of invalid type
			def initialize(dim)
				@dim = dim
				if block_given?
					@data = []
					dim.times do |d|
						val = yield(d)
						raise(KXI::Exceptions::InvalidTypeException.new(val.type, Numeric)) unless val.is_a?(Numeric)
						@data.push(val.to_f)
					end
				else
					@data = [0.0] * dim
				end
			end
			
			def str(num)
				return (num == num.to_i ? num.to_i : num).to_s
			end
			
			# Gets the value of vector at specific dimension
			# @param [Integer] idx Dimension to return
			# @raise [KXI::Exceptions::OutOfRangeException] When dimension is out of range
			# @return [Numeric] Value of vector at given dimension
			def [](idx)
				raise(KXI::Exceptions::OutOfRangeException.new(idx, 0, @dim - 1)) if idx < 0 or idx >= @dim
				raise(KXI::Exceptions::InvalidTypeException.new(value.type, Numeric)) unless value.is_a?(Numeric)
				return @data[idx]
			end
			
			# Sets the value of vector at specific dimension
			# @overload []=(idx, val)
			# 	Sets the value of vector at specific dimension
			# 	@param [Integer] idx Dimension to set
			# 	@param [Integer] val Value to set the dimension to
			# 	@raise [KXI::Exceptions::OutOfRangeException] When dimension is out of range
			# 	@raise [KXI::Exceptions::InvalidTypeException] When value has invalid type
			# 	@return [Numeric] Value passed to function
			# @overload []=(idx, val)
			# 	Sets the values of vector to range of values starting from specific dimension
			# 	@param [Integer] idx Dimension to start at
			# 	@param [Array] val Values to set the dimensions to
			# 	@raise [KXI::Exceptions::OutOfRangeException] When dimension is out of range
			# 	@raise [KXI::Exceptions::InvalidTypeException] When value has invalid type
			# 	@return [Numeric] Value passed to function
			def []=(idx, val)
				raise(KXI::Exceptions::OutOfRangeException.new(idx, 0, @dim - 1)) if idx < 0 or idx >= @dim
				if val.is_a?(Numeric)
					@data[idx] = val.to_f
				elsif val.is_a?(Array)
					i = 0
					while idx + i < @dim and i < val.length
						v = val[i]
						raise(KXI::Exceptions::InvalidTypeException.new(v.type, Numeric)) unless v.is_a?(Numeric)
						@data[idx + i] = v.to_f
					end
				else
					raise(KXI::Exceptions::InvalidTypeException.new(val.type, Numeric, Array))
				end
				return val
			end
			
			# Iterates over each element of vector
			# @yield [d, val] Iterator function
			# @yieldparam [Integer] d Dimension passed to iterator
			# @yieldparam [Numeric] val Value at that specific dimension
			def each
				return unless block_given?
				@dim.times do |d|
					yield(d, @data[d])
				end
			end
			
			# Dot multiplies vectors
			# @param [KXI::Math::Vector] vec Vector to multiply with
			# @return [Numeric] Dot product of vectors
			def dot(vec)
				raise(KXI::Exception::DimensionMismatchException.new(vec.dimension, dimension)) unless vec.dimension == dimension
				sum = 0
				@dim.times { |d| sum += @data[d] * vec[d] }
				return sum
			end
			
			# Converts vector to string
			# @param [Boolean] d Determines whether string should be decorated
			# @return [String] String representation of vector
			def to_s(d = true)
				just = 0
				@dim.times { |m| len = str(@data[m]).length; just = len if len > just }
				ret = ''
				@dim.times do |m|
					ret += $/ if m > 0
					ret += '|' if d
					ret += str(@data[m]).rjust(just, ' ')
					ret += '|' if d
				end
				return ret
			end
			
			# Adds vectors
			# @param [Vector] other Vector to add
			# @raise [KXI::Exceptions::InvalidTypeException] When type of value is invalid
			# @return [KXI::Math::Vector] New vector equivalent to the sum of vectors
			def +(other)
				if other.is_a?(KXI::Math::Vector)
					return KXI::Math::Vector.new(@dim) { |d| @data[d] + other[d] }
				else
					raise(KXI::Exceptions::InvalidTypeException.new(other.class, Numeric))
				end
			end
			
			# Subtracts vectors
			# @param [Vector] other Vector to subtract
			# @raise [KXI::Exceptions::InvalidTypeException] When type of value is invalid
			# @return [KXI::Math::Vector] New vector equivalent to the difference of vectors
			def -(other)
				if other.is_a?(KXI::Math::Vector)
					return KXI::Math::Vector.new(@dim) { |d| @data[d] - other[d] }
				else
					raise(KXI::Exceptions::InvalidTypeException.new(other.class, Numeric))
				end
			end
			
			# Multiplies the vector with -1
			# @return [KXI::Math::Vector] New opposite vector
			def -@
				return KXI::Math::Vector.new(@dim) { |d| -@data[d] }
			end
			
			# Multiplies vector
			# @overload *(other)
			# 	Preforms scalar-vector multiplication
			# 	@param [Numeric] other Scalar to multiply with
			# 	@raise [KXI::Exceptions::InvalidTypeException] When other value is of invalid type
			# 	@return [KXI::Math::Vector] New vector equivalent to the scaled vector
			# @overload *(other)
			# 	Preforms vector-vector multiplication
			# 	@param [Numeric] other Vector to multiply with
			# 	@raise [KXI::Exceptions::InvalidTypeException] When other value is of invalid type
			# 	@return [KXI::Math::Vector] New vector equivalent to the multiplied vector
			def *(other)
				if other.is_a?(Numeric)
					return KXI::Math::Vector.new(@dim) { |d| other * @data[d] }
				elsif other.is_a?(KXI::Math::Vector)
					return KXI::Math::Vector.new(@dim) { |d| @data[d] * other[d] }
				elsif raise(KXI::Exceptions::InvalidTypeException.new(other.class, Numeric, KXI::Math::Vector))
				end
			end
			
			# Compares vector
			# @param [void] other Value to compare to
			# @return [Boolean] True if vector is equivalent to given value; false otherwise
			def ==(other)
				return false unless other.is_a?(KXI::Math::Vector)
				return false if @dim != other.dimension
				@dim.times do |d|
					return false if @data[d] != other[d]
				end
				return true
			end
			
			private :str
		end
	end
end